(* file: queens.ml
 * author: Bob Muller
 *
 * CS1103 Computer Science I Honors
 *
 * This program solves the 8-queens problem. It uses Asai's Universe library.
 *
 * Compile and run:
 *
 * > make
 *> ./go
 *)
open World
open Color
open Image
open Cs1103

let imageSide = 100.               (* determined by size of crown.png *)
let boardSize = 8. *. imageSide

let emptyBoardImage = Image.rectangle boardSize boardSize Color.white
let whiteSquare = Image.rectangle imageSide imageSide Color.darkGray
let graySquare  = Image.rectangle imageSide imageSide Color.gray20
let crown = Image.read_image "crown.png"

type piece  = Queen | Empty
type board = (piece array) array
type position = {row : int; col : int}
type memory = position Stack.t

type model = {board : board;
              position : position;
              memory : memory;
              solved : bool}

(* next : position -> position
*)
let next position =                 (* starts at {row=0; col=0}; returns *)
  if position.col < 7 then          (* {row=8; col=0} when all done.     *)
    {row = position.row; col = position.col + 1}
  else
    {row = position.row + 1; col = 0}

(* attacksFromAbove : position -> position -> board -> bool
*)
let attacksFromAbove pos1 pos2 board =
  let hasAQueen  = board.(pos1.row).(pos1.col) = Queen in
  let sameColumn = pos1.col = pos2.col in
  let onDiagonal = (abs (pos1.row - pos2.row)) = (abs (pos1.col - pos2.col))
  in
  hasAQueen && (sameColumn || onDiagonal)

(* attackedFromAbove : position -> board -> bool
*)
let attackedFromAbove position board =
  let rec repeat position' =
    match position'.row = position.row with
    | true  -> false
    | false ->
      (match attacksFromAbove position' position board with
		   | true  -> true
       | false -> repeat (next position'))
  in
  repeat {row = 0; col = 0}

(* update : model -> model World.t
*)
let update ({board; memory; position} as model) =
  let rec solve position =
    match position.row = 8 with
    | true  -> World.World({ model with solved = true})
    | false ->
      (match (attackedFromAbove position board) with
       | true  ->
         (match position.col < 7 with
          | true  -> solve {position with col = position.col + 1}
          | false -> backTrack model)
       | false ->
         board.(position.row).(position.col) <- Queen;
         Stack.push position memory;
         World.World({ model with position = {row = position.row + 1; col=0}}))

  and backTrack model =
    match Stack.is_empty model.memory with
    | true  -> failwith "queens: there is no solution.\n"
    | false ->
      let position' = Stack.pop model.memory
	    in
	    board.(position'.row).(position'.col) <- Empty ;
	    (match position'.col = 7 with
       | true  -> backTrack model               (* keep popping ... *)
       | false -> solve (next position'))
  in
  solve position

(* view : model -> Image.t
*)
let view {board; memory} =
  let getImage piece row col =
    match piece with
    | Empty -> (if ((row + col) mod 2) = 0 then whiteSquare else graySquare)
    | Queen -> crown in
  let rec makeBoardImage position image =
    let (row, col) = (position.row, position.col) in
    match row = 8 with
    | true  -> image
    | false ->
      let img = getImage board.(row).(col) row col in
	    let (x, y) = ((i2F col) *. imageSide, ((i2F row) *. imageSide)) in
	    let newImage = Image.place_image img (x, y) image
	    in
      makeBoardImage (next position) newImage
  in
  makeBoardImage {row = 0; col = 0} emptyBoardImage

let finished world = world.solved

let queens () =
  let initialModel = {board = Array.make_matrix 8 8 Empty;
                      position = {row = 0; col = 0};
                      memory = Stack.create ();
                      solved = false}
  in
  World.big_bang
    initialModel
    ~name: "Eight Queens!"
    ~width: (f2I boardSize)
    ~height: (f2I boardSize)
		 ~on_tick: update
		 ~rate: 1.0
		 ~to_draw: view
		 ~stop_when: finished
		 ~to_draw_last: view

let () = queens ()
